/*
 * File: cursor.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 */


/***

The routines in this file move the cursor around on the screen. They
compute a new value for the cursor, then adjust cursor. The display code
always updates the cursor location, so only moves between lines, or
functions that adjust the top line in the window and invalidate the
framing, are hard.  Cursor is refered to as "dot" throughout.

***/

#include	<stdio.h>
#include	<ctype.h>
#include	"rho.h"
#include	"cursor.h"
#include	"cmode.h"
#include	"messages.h"
#include	"utils.h"
#include	"word.h"


/*++

Dot to beginning of current line.

++*/
int beginning_of_line()
{
    curwp->w_doto = 0;
    return (TRUE);
}


/*++

Move dot backward n chars. If n < 0, call next_character() to do the
move. Else compute new dot location. Error if move exits buffer
(enters header line). Set flag if line pointer for dot changes.

++*/
int previous_character(f, n)
int f, n;

{
    LINE *lp;

    if (n < 0)
        return (next_character(f, -n));
    while (n--) {
        if (curwp->w_doto == 0) {
            if ((lp = lback(curwp->w_dotp)) == curbp->b_linep)
                return (FALSE);
            curwp->w_dotp = lp;
            curwp->w_doto = lused(lp);
            curwp->w_flag |= WFMOVE;
        } else
            curwp->w_doto--;
    }
    return (TRUE);
}


/*++

Dot to end of current line.

++*/
int end_of_line()
{
    curwp->w_doto = lused(curwp->w_dotp);
    return (TRUE);
}


/*++

Move dot forward n chars. If n < 0 call previous_character() to do
the move. Else compute new dot location, and move dot. Error if move
exits buffer. Set WFMOVE flag if dot's line changes.

++*/
int next_character(f, n)
int f, n;

{
    if (n < 0)
        return (previous_character(f, -n));
    while (n--) {
        if (curwp->w_doto == lused(curwp->w_dotp)) {
            if (curwp->w_dotp == curbp->b_linep)
                return (FALSE);
            curwp->w_dotp = lforw(curwp->w_dotp);
            curwp->w_doto = 0;
            curwp->w_flag |= WFMOVE;
        } else
            curwp->w_doto++;
    }
    return (TRUE);
}


/*++

Move dot to a given line.
Argument (n) must be positive integer.

++*/
int goto_line(f, n)
int f, n;

{

    int status;                 /* status return */
    char arg[NSTRING];          /* buffer to hold argument */

    if (f == FALSE) {
        if ((status = mlreply(TEXT7, arg, NSTRING)) != TRUE) {  /*"Line to GOTO:" */
            mlwrite(TEXT8);     /*"[Aborted]" */
            return (status);
        }
        n = str2int(arg);
    }

    if (n < 1)
        return (FALSE);

    curwp->w_dotp = lforw(curbp->b_linep);      /*  dot row = line after hdr line  */
    curwp->w_doto = 0;          /*  dot col = 1  */
    return (next_line(f, n - 1));
}


/*++

Move dot to beginning of buffer. Always set WFMOVE flag.

++*/
int beginning_of_buffer()
{
    curwp->w_dotp = lforw(curbp->b_linep);
    curwp->w_doto = 0;
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}


/*++

Move dot to the end of the buffer. 

++*/
int end_of_buffer()
{
    curwp->w_dotp = curbp->b_linep;
    curwp->w_doto = 0;
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}



/*++

This routine, given a pointer to a LINE, and the current dot goal
column, return the best choice for the offset. The offset is returned.
Used by next|previous_line()

++*/

static int get_line_goal(dlp)
LINE *dlp;

{
    int c;
    int col;
    int newcol;
    int dbo;

    col = 0;
    dbo = 0;
    while (dbo != lused(dlp)) {
        c = lgetc(dlp, dbo);
        newcol = col;
        if (c == '\t')
            newcol += -(newcol % hardtab) + (hardtab - 1);
        else if (c < 0x20 || c == 0x7F)
            ++newcol;
        ++newcol;
        if (newcol > curgoal)
            break;
        col = newcol;
        ++dbo;
    }
    return (dbo);
}


/*++

Move dot forward by full lines. If n < 0, call the previous_line function. The
last command controls how the goal column is set.

++*/
int next_line(f, n)
int f, n;

{
    LINE *dlp;

    if (n < 0)
        return (previous_line(f, -n));

    /* if we are on the last line as we start, return */
    if (curwp->w_dotp == curbp->b_linep)
        return (FALSE);

    /* if the last command was not a line move, reset the goal column */
    if ((lastflag & CFCPCN) == 0)
        curgoal = getccol(FALSE);

    /* flag this command as a line move */
    thisflag |= CFCPCN;

    /* and move the point down */
    dlp = curwp->w_dotp;
    while (n-- && dlp != curbp->b_linep)
        dlp = lforw(dlp);

    /* reseting the current position */
    curwp->w_dotp = dlp;
    curwp->w_doto = get_line_goal(dlp);
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}


/*++

Like "next_line", but goes backwards. 

++*/
int previous_line(f, n)
int f, n;

{
    LINE *dlp;

    if (n < 0)
        return (next_line(f, -n));


    /* if we are on the first line as we start, fail */
    if (lback(curwp->w_dotp) == curbp->b_linep)
        return (FALSE);

    /* if the last command was not a line move, reset the goal column */
    if ((lastflag & CFCPCN) == 0)
        curgoal = getccol(FALSE);

    /* flag this command as a line move */
    thisflag |= CFCPCN;

    /* and move the point up */
    dlp = curwp->w_dotp;
    while (n-- && lback(dlp) != curbp->b_linep)
        dlp = lback(dlp);

    /* reseting the current position */
    curwp->w_dotp = dlp;
    curwp->w_doto = get_line_goal(dlp);
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}


/*++

Move dot to the next paragraph-beginning behind the dot.
We look for a <NL><NL> or <NL><TAB> or <NL><SPACE>

++*/
int previous_paragraph(f, n)
int f, n;

{
    int suc;                    /* success of last previous_character */
    int found;                  /* found non space character */

    if (n < 0)                  /* the other way... */
        return (next_paragraph(f, -n));

    while (n-- > 0) {           /* for each one asked for */

        /* first scan back until we are in a word */
        suc = previous_character(FALSE, 1);
        while (!inword() && suc)
            suc = previous_character(FALSE, 1);
        curwp->w_doto = 0;      /* and go to the B-O-Line */

        /* and scan back until we hit a target pattern */
        while (lback(curwp->w_dotp) != curbp->b_linep)
            if (lused(curwp->w_dotp) != 0) {
                found = 0;
                for (suc = 0; suc < lused(curwp->w_dotp); suc++)
                    if (!isspace
                        (lgetc(curwp->w_dotp, suc + curwp->w_doto))) {
                        found = 1;
                        break;
                    }
                if (found == 0)
                    break;
                else
                    curwp->w_dotp = lback(curwp->w_dotp);
            } else
                break;

        /* and then forward until we are in a word */
        suc = next_character(FALSE, 1);
        while (suc && !inword())
            suc = next_character(FALSE, 1);
    }
    curwp->w_flag |= WFMOVE;    /* force screen update */
    return (TRUE);
}


/*++

Move dot to the next paragraph-ending in front of the dot.
Otherwise, as for previous_paragraph.

++*/
int next_paragraph(f, n)
int f, n;

{
    int suc;                    /* success of last previous_character */
    int found;                  /* found non space character */

    if (n < 0)
        return (previous_paragraph(f, -n));

    while (n-- > 0) {           /* for each one asked for */

        /* first scan forward until we are in a word */
        suc = next_character(FALSE, 1);
        while (!inword() && suc)
            suc = next_character(FALSE, 1);
        curwp->w_doto = 0;      /* and go to the B-O-Line */
        if (suc)                /* of next line if not at EOF */
            curwp->w_dotp = lforw(curwp->w_dotp);

        /* and scan forword until we hit a <NL><NL> or <NL><TAB>
           or a <NL><SPACE>                                     */
        while (curwp->w_dotp != curbp->b_linep) {
            if (lused(curwp->w_dotp) != 0) {
                found = 0;
                for (suc = 0; suc < lused(curwp->w_dotp); suc++)
                    if (!isspace
                        (lgetc(curwp->w_dotp, suc + curwp->w_doto))) {
                        found = 1;
                        break;
                    }
                if (found == 0)
                    break;
                else
                    curwp->w_dotp = lforw(curwp->w_dotp);
            } else
                break;
        }
        /* and then backward until we are in a word */
        suc = previous_character(FALSE, 1);
        while (suc && !inword()) {
            suc = previous_character(FALSE, 1);
        }
        curwp->w_doto = lused(curwp->w_dotp);   /* and to the EOL */
    }
    curwp->w_flag |= WFMOVE;    /* force screen update */
    return (TRUE);
}


/*++

Scroll forward by a full page. The overlap in the arithmetic on the
window size is overlap between screens. Because this zaps the top line
in the window, we have to do a hard update.

++*/
int page_down()
{
    LINE *lp;
    int n;

    n = curwp->w_ntrows - overlap;      /* Default scroll.      */
    if (n <= 0)                 /* Forget the overlap   */
        n = 1;                  /* if tiny window.      */

    lp = curwp->w_linep;
    while (n-- && lp != curbp->b_linep)
        lp = lforw(lp);
    curwp->w_linep = lp;
    curwp->w_dotp = lp;
    curwp->w_doto = 0;
    curwp->w_flag |= WFHARD;
    return (TRUE);
}


/*++

Like "page_down", but in reverse

++*/
int page_up()
{
    LINE *lp;
    int n;

    n = curwp->w_ntrows - overlap;      /* Default scroll.      */
    if (n <= 0)                 /* Don't blow up if the */
        n = 1;                  /* window is tiny.      */

    lp = curwp->w_linep;
    while (n-- && lback(lp) != curbp->b_linep)
        lp = lback(lp);
    curwp->w_linep = lp;
    curwp->w_dotp = lp;
    curwp->w_doto = 0;
    curwp->w_flag |= WFHARD;
    return (TRUE);
}


/*++

Set the mark in the current window to the value of dot in the window. 

++*/
int set_mark(f, n)
int f, n;

{
    /* make sure it is in range */
    if (f == FALSE)
        n = 0;
    n %= NMARKS;

    curwp->w_markp[n] = curwp->w_dotp;
    curwp->w_marko[n] = curwp->w_doto;
    mlwrite(TEXT9, n);          /* "[Mark %d set]" */
    return (TRUE);
}


/*++

Swap the values of dot and "mark" in the current window.
The only possible error is "no mark".

++*/
int swap_mark(f, n)
int f, n;

{
    LINE *odotp;
    int odoto;

    /* make sure it is in range */
    if (f == FALSE)
        n = 0;
    n %= NMARKS;

    if (curwp->w_markp[n] == NULL) {
        mlwrite(TEXT11, n);     /*"No mark %d in this window" */
        return (FALSE);
    }
    odotp = curwp->w_dotp;
    odoto = curwp->w_doto;
    curwp->w_dotp = curwp->w_markp[n];
    curwp->w_doto = curwp->w_marko[n];
    curwp->w_markp[n] = odotp;
    curwp->w_marko[n] = odoto;
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}



/*++

Go to a mark in the current window. The only possible error is "no mark".
NOTE: Set_mark can set n <= NMARKS mark but nothing is using set_mark
beyond mark 0.  That makes this a partial swap_mark...

++*/
int goto_mark(f, n)
int f, n;

{
    /* make sure it is in range */
    if (f == FALSE)
        n = 0;
    n %= NMARKS;

    if (curwp->w_markp[n] == NULL) {
        mlwrite(TEXT11, n);     /* "No mark %d in this window" */
        return (FALSE);
    }
    curwp->w_dotp = curwp->w_markp[n];
    curwp->w_doto = curwp->w_marko[n];
    curwp->w_flag |= WFMOVE;
    return (TRUE);
}



/*++

Display the current position of the cursor, in origin 1 X-Y coordinates,
the character that is under the cursor (in hex), and the fraction of the
text that is before the cursor. The displayed column is not the current
column, but the column that would be used on an infinite width display.

++*/
int cursor_position(f, n)
int f, n;
{
    LINE *lp;                   /* current line */
    long numchars;              /* # of chars in file */
    long numlines;              /* # of lines in file */
    long predchars;             /* # chars preceding point */
    long predlines;             /* # lines preceding point */
    int curchar;                /* character under cursor */
    int ratio;
    int col;
    int savepos;                /* temp save for current offset */
    int ecol;                   /* column pos/end of current line */

    f = n = 0;                  /* splint */

    /* starting at the beginning of the buffer */
    lp = lforw(curbp->b_linep);
    curchar = '\r';

    /* start counting chars and lines */
    numchars = 0;
    numlines = 0L;
    while (lp != curbp->b_linep) {
        /* if we are on the current line, record it */
        if (lp == curwp->w_dotp) {
            predlines = numlines;
            predchars = numchars + curwp->w_doto;
            if ((curwp->w_doto) == lused(lp))
                curchar = '\r';
            else
                curchar = lgetc(lp, curwp->w_doto);
        }
        /* on to the next line */
        ++numlines;
        numchars += lused(lp) + 1;
        lp = lforw(lp);
    }

    /* if at end of file, record it */
    if (curwp->w_dotp == curbp->b_linep) {
        predlines = numlines;
        predchars = numchars;
    }

    /* Get real column and end-of-line column. */
    col = getccol(FALSE);
    savepos = curwp->w_doto;
    curwp->w_doto = lused(curwp->w_dotp);
    ecol = getccol(FALSE);
    curwp->w_doto = savepos;

    ratio = 0;                  /* Ratio before dot. */
    if (numchars != 0)
        ratio = (100L * predchars) / numchars;

    /* summarize and report the info */

    mlwrite(TEXT60, predlines + 1, numlines + 1, col + 1, ecol, predchars, numchars, ratio, curchar);   /*     "Line Col Char char = 0x" */
    return (TRUE);
}
