/*
 * File: exec.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */

/***

This file is the heart of the editor.  All the user requests for
action come here.  They are then routed to "some-command" linked to a
function in rho_names.h (the functions are in various .c files) or to
the macro execution of a file, buffer, or stored procedure. The function
execute_buffer() in this file is the actual engine for this latter,
as all macro execution is parsed in a buffer.  From execute_buffer,
these are then routed down through:

    macros.c - parses and routes the tokens on a line of the buffer,
               routes the gets and sets of envars and uvars.
    envars.c - handles the gets and sets of its envars, initialized
               in main.h's names[]
     ufunc.c - handles the execution its user functions (documented
               in help4rho.txt) and the gets and sets of user-created
               variables or uvars.                   

NOTE: In the original uemacs, a "macro" was considered to be one of
40 stored buffers of data, stored by "store-macro" and executed by
"exec.cbuf()"  What I have put in macros.c is the top level of parsing
after the execute_buffer is called.  This level was identified by
tokens of the (regex) form TK[.*]
              
***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"exec.h"
#include	"bind.h"
#include	"buffer.h"
#include	"display.h"
#include	"file.h"
#include	"fileio.h"
#include	"line.h"
#include	"macros.h"
#include	"messages.h"
#include	"utils.h"



/*++

Free a list of while block pointers.
Local helper.

++*/
static void freewhile(wp)
WHBLOCK *wp;                    /* head of structure to free */

{
    if (wp != NULL) {
        freewhile(wp->w_next);
        free((char *) wp);
    }
}



/*++

Display a macro execution error along with the buffer and line
currently being executed.

++*/
static void error_msg(mesg, bp, lp)
char *mesg;                     /* error message to display */
BUFFER *bp;                     /* buffer error occured in  */
LINE *lp;                       /* line error occured on    */

{
    char buf[NSTRING];

    exec_error = TRUE;

    /* build error message line */
    strcpy(buf, "\n");
    strcat(buf, mesg);
    strcat(buf, TEXT229);       /*     " in < " */
    strcat(buf, bp->b_name);
    strcat(buf, TEXT230);       /*     "> at line " */
    strcat(buf, int2str(getlinenum(bp, lp)));
    mlforce(buf);
}




/*++

Takes either a message-line or the current line in a buffer.  In macros,
the buffer is unseen of course. Global helper.

++*/
int next_arg(prompt, buffer, size, terminator)
char *prompt;                   /* prompt to use if we must be interactive */
char *buffer;                   /* buffer to put token into */
int size;                       /* size of the buffer */
int terminator;                 /* terminating char to be used on interactive fetch */

{
    char *sp;                   /* return pointer from get_TKval() */

    /* if we are interactive, ask */
    if (clexec == FALSE) {
        /* prompt the user for the input string */
        if (discmd) {
            if (prompt)
                mlwrite(prompt);
        } else
            movecursor(term.t_nrow, 0);

        return (mlgetstr(buffer, size, terminator));
    }

    /* grab token and advance past */
    execstr = next_token(execstr, buffer, size);

    /* evaluate it */
    if ((sp = get_TKval(buffer)) == NULL)
        return (FALSE);
    strcpy(buffer, sp);
    return (TRUE);
}


/*++

Method to execute cmd or macro from the message line.

++*/
int execute_command(f, n)
int f, n;                       /* command arguments, passed through to command executed */

{
    char buffer[NSTRING];       /* buffer to store function name */
    int (*kfunc) ();            /* ptr to the function to execute */
    int status;

    /* if we are non-interactive, force the command interactively */
    if (clexec == TRUE) {
        /* grab token and advance past */
        execstr = next_token(execstr, buffer, NPAT);

        /* evaluate it */
        strcpy(buffer, fixnull(get_TKval(buffer)));
        if (strcmp(buffer, errorm) == 0)
            return (FALSE);

        /* and look it up */
        if ((kfunc = get_fn_name(buffer)) == NULL) {
            mlwrite(TEXT16);    /* "[No such Function]" */
            return (FALSE);
        }

        /* and execute it as if interactive */
        clexec = FALSE;
        status = (*kfunc) (f, n);       /* call the function */
        clexec = TRUE;
        return (status);
    }

    /* prompt the user to type a named command */
    /* and get the function name to execute    */
    kfunc = prompt4command(": ");
    if (kfunc == NULL) {
        mlwrite(TEXT16);        /* "[No such function]" */
        return (FALSE);
    }

    /* and then execute the command */
    return ((*kfunc) (f, n));
}



/*++

Pull the next line in the macro.  A kind of wrapper for next_arg.
Global helper.

++*/
int get_macro_line(token)
char *token;                    /* buffer to place argument */

{
    int savcle;                 /* buffer to store original clexec */
    int status;

    savcle = clexec;            /* save execution mode */
    clexec = TRUE;              /* get the argument */
    status = next_arg("", token, NSTRING, ctoec('\r'));
    clexec = savcle;            /* restore execution mode */
    return (status);
}


/*  circular reference in do_cmd <-> do_buffer   */
static int do_cmd(char *);

/*++

        Execute the contents of the buffer pointed to by the passed BP.
        The is the heart of all macro execution.

	Directives start with a "!" and include:

	!endm		End a macro
	!if (cond)	conditional execution
	!else
	!endif
	!return		Return (terminating current macro)
	!goto <label>	Jump to a label in the current macro
	!force		Force macro to continue...even if command fails
	!while (cond)	Execute a loop if the condition is true
	!endwhile
	
	Line Labels begin with a "*" as the first nonblank char, like:

	*LBL01
++*/
int do_buffer(bp)
BUFFER *bp;                     /* buffer to execute */

{
    int status;                 /* status return */
    LINE *lp;                   /* pointer to line to execute */
    LINE *hlp;                  /* pointer to line header */
    LINE *glp;                  /* line to goto */
    LINE *mp;                   /* Macro line storage temp */
    int dirnum;                 /* directive index */
    size_t linlen;              /* length of line to execute */
    size_t i;                   /* index */
    int force;                  /* force TRUE result? */
    WNDW *wp;                   /* ptr to windows to scan */
    WHBLOCK *whlist;            /* ptr to !WHILE list */
    WHBLOCK *scanner;           /* ptr during scan */
    WHBLOCK *whtemp;            /* temporary ptr to a WHBLOCK */
    char *einit;                /* initial value of eline */
    char *eline;                /* text of line to execute */
    char tkn[NSTRING];          /* buffer to evaluate an expresion in */


    /* clear IF level flags/while ptr */
    execlevel = 0;
    whlist = NULL;
    scanner = NULL;

    /* flag we are executing the buffer */
    bp->b_exec += 1;

    /* scan the buffer to execute, building WHILE header blocks */
    hlp = bp->b_linep;
    lp = lforw(hlp);
    while (lp != hlp) {

        /* scan the current line */
        eline = ltext(lp);
        i = lused(lp);

        /* trim leading whitespace */
        while (i-- > 0 && (*eline == ' ' || *eline == '\t'))
            ++eline;

        /* if theres nothing here, don't bother */
        if (i <= 0)
            goto nxtscan;

        /* if is a while directive, make a block... */
        if (eline[0] == '!' && eline[1] == 'w' && eline[2] == 'h') {
            whtemp = (WHBLOCK *) malloc(sizeof(WHBLOCK));
            whtemp->w_begin = lp;
            whtemp->w_type = BTWHILE;
            whtemp->w_next = scanner;
            scanner = whtemp;
        }

        /* if is a BREAK directive, make a block... */
        if (eline[0] == '!' && eline[1] == 'b' && eline[2] == 'r') {
            if (scanner == NULL) {
                error_msg(TEXT120, bp, lp);     /*"%%!BREAK outside of any !WHILE loop" */
                goto eabort;
            }
            whtemp = (WHBLOCK *) malloc(sizeof(WHBLOCK));
            whtemp->w_begin = lp;
            whtemp->w_type = BTBREAK;
            whtemp->w_next = scanner;
            scanner = whtemp;
        }

        /* if it is an endwhile directive, record the spot... */
        if (eline[0] == '!' && strncmp(&eline[1], "endw", 4) == 0) {
            if (scanner == NULL) {
                error_msg(TEXT121, bp, lp);     /*"%%!ENDWHILE with no preceding !WHILE" */
                goto eabort;
            }
            /* move top records from the scanner list to the
               whlist until we have moved all BREAK records
               and one WHILE record */
            do {
                scanner->w_end = lp;
                whtemp = whlist;
                whlist = scanner;
                scanner = scanner->w_next;
                whlist->w_next = whtemp;
            }
            while (whlist->w_type == BTBREAK);
        }

      nxtscan:                 /* on to the next line */
        lp = lforw(lp);
    }

    /* while and endwhile should match! */
    if (scanner != NULL) {
        error_msg(TEXT122, bp, lp);     /*"%%!WHILE with no matching !ENDWHILE" */
        goto eabort;
    }

    /* let the first command inherit the flags from the last one.. */
    thisflag = lastflag;

    /* starting at the beginning of the buffer */
    hlp = bp->b_linep;
    lp = lforw(hlp);
    while (lp != hlp && eexitflag == FALSE) {

        /* allocate eline and copy macro line to it */
        linlen = lused(lp);
        einit = eline = malloc(linlen + 1);
        strncpy(eline, ltext(lp), linlen);
        eline[linlen] = 0;      /* make sure it ends */

        /* trim leading whitespace */
        while (*eline == ' ' || *eline == '\t')
            ++eline;

        /* dump comments and blank lines */
        if (*eline == ';' || *eline == 0)
            goto onward;

        /* Parse directives here.... */
        dirnum = -1;
        if (*eline == '!') {
            /* Find out which directive this is */
            ++eline;
            for (dirnum = 0; dirnum < NUMDIRS; dirnum++)
                if (strncmp
                    (eline, dname[dirnum],
                     (size_t) dname_len[dirnum]) == 0)
                    break;

            /* and bitch if it's illegal */
            if (dirnum == NUMDIRS) {
                error_msg(TEXT124, bp, lp);     /*"%%Unknown Directive" */
                goto eabort;
            }

            /* service only the !ENDM macro here */
            if (dirnum == DENDM) {
                mstore = FALSE;
                bstore = NULL;
                goto onward;
            }

            /* restore the original eline.... */
            --eline;
        }

        /* if macro store is on, just salt this away */
        if (mstore) {
            /* allocate the space for the line */
            linlen = strlen(eline);
            if ((mp = alloc_line(linlen)) == NULL)
                goto eabort;

            /* copy the text into the new line */
            for (i = 0; i < linlen; ++i)
                lputc(mp, i, eline[i]);

            /* attach the line to the end of the buffer */
            bstore->b_linep->l_bp->l_fp = mp;
            mp->l_bp = bstore->b_linep->l_bp;
            bstore->b_linep->l_bp = mp;
            mp->l_fp = bstore->b_linep;
            goto onward;
        }

        force = FALSE;

        /* dump comments */
        if (*eline == '*')
            goto onward;

        /* now, execute directives */
        if (dirnum != -1) {
            /* skip past the directive */
            while (*eline && *eline != ' ' && *eline != '\t')
                ++eline;
            execstr = eline;

            switch (dirnum) {
            case DIF:          /* IF directive */
                /* grab the value of the logical exp */
                if (execlevel == 0) {
                    if (get_macro_line(tkn) != TRUE) {
                        free(einit);
                        goto eexec;
                    }
                    if (str2bool(tkn) == FALSE)
                        ++execlevel;
                } else
                    ++execlevel;
                goto onward;

            case DWHILE:       /* WHILE directive */
                /* grab the value of the logical exp */
                if (execlevel == 0) {
                    if (get_macro_line(tkn) != TRUE) {
                        free(einit);
                        goto eexec;
                    }
                    if (str2bool(tkn) == TRUE)
                        goto onward;
                }
                /* drop down and act just like !BREAK */

            case DBREAK:       /* BREAK directive */
                if (dirnum == DBREAK && execlevel)
                    goto onward;

                /* jump down to the endwhile */
                /* find the right while loop */
                whtemp = whlist;
                while (whtemp) {
                    if (whtemp->w_begin == lp)
                        break;
                    whtemp = whtemp->w_next;
                }

                if (whtemp == NULL) {
                    error_msg(TEXT126, bp, lp); /*"%%Internal While loop error" */
                    goto eabort;
                }

                /* reset the line pointer back.. */
                lp = whtemp->w_end;
                goto onward;

            case DELSE:        /* ELSE directive */
                if (execlevel == 1)
                    --execlevel;
                else if (execlevel == 0)
                    ++execlevel;
                goto onward;

            case DENDIF:       /* ENDIF directive */
                if (execlevel)
                    --execlevel;
                goto onward;

            case DGOTO:        /* GOTO directive */
                /* only if we are currently executing */
                if (execlevel == 0) {

                    /* grab label to jump to */
                    eline = next_token(eline, golabel, NPAT);
                    linlen = strlen(golabel);
                    glp = lforw(hlp);
                    while (glp != hlp) {
                        if ((lused(glp) >= linlen) &&
                            (lgetc(glp, 0) == '*') &&
                            (strncmp(((char *) ltext(glp)) + 1,
                                     golabel, linlen) == 0)) {
                            lp = glp;
                            goto onward;
                        }
                        glp = lforw(glp);
                    }
                    error_msg(TEXT127, bp, lp); /*"%%No such label" */
                    goto eabort;
                }
                goto onward;

            case DRETURN:      /* RETURN directive */
                if (execlevel == 0) {
                    free(einit);
                    goto eexec;
                }
                goto onward;

            case DENDWHILE:    /* ENDWHILE directive */
                if (execlevel) {
                    --execlevel;
                    goto onward;
                } else {
                    /* find the right while loop */
                    whtemp = whlist;
                    while (whtemp) {
                        if (whtemp->w_type == BTWHILE
                            && whtemp->w_end == lp)
                            break;
                        whtemp = whtemp->w_next;
                    }

                    if (whtemp == NULL) {
                        error_msg(TEXT126, bp, lp);     /*"%%Internal While loop error" */
                        goto eabort;
                    }

                    /* reset the line pointer back.. */
                    lp = lback(whtemp->w_begin);
                    goto onward;
                }

            case DFORCE:       /* FORCE directive */
                force = TRUE;

            }
        }

        /* execute the statement */
        status = do_cmd(eline);
        if (force)              /* force the status */
            status = TRUE;

        /* check for a command error */
        if (status != TRUE) {

            /* look if buffer is showing */
            wp = wheadp;
            while (wp != NULL) {
                if (wp->w_bufp == bp) {
                    /* and point it */
                    wp->w_dotp = lp;
                    wp->w_doto = 0;
                    wp->w_flag |= WFHARD;
                }
                wp = wp->w_wndp;
            }

            /* in any case set the buffer . */
            bp->b_dotp = lp;
            bp->b_doto = 0;

            error_msg(TEXT219, bp, lp); /* "%%Macro Failed" */

            execlevel = 0;
            freewhile(whlist);
            bp->b_exec -= 1;
            free(einit);
            return (status);
        }

      onward:                  /* on to the next line */
        free(einit);
        lp = lforw(lp);

    }

  eexec:                       /* exit the current function */
    execlevel = 0;
    freewhile(whlist);
    bp->b_exec -= 1;
    return (TRUE);

  eabort:                      /* exit the current function with a failure */
    execlevel = 0;
    freewhile(whlist);
    bp->b_exec -= 1;
    free(einit);
    return (FALSE);
}


/*++

Chop a token off a string and return a pointer past the ex-token.
Global helper.

++*/
char *next_token(src, tok, size)
char *src, *tok;                /* source string, destination token string */
int size;                       /* maximum size of token */

{
    int quotef;                 /* is the current string quoted? */
    char c;                     /* temporary character */

    /* first scan past any whitespace in the source string */
    while (*src == ' ' || *src == '\t')
        ++src;

    /* scan through the source string */
    quotef = FALSE;
    while (*src) {
        /* process special characters */
        if (*src == '~') {
            ++src;
            if (*src == 0)
                break;
            switch (*src++) {
            case 'r':
                c = 13;
                break;
            case 'n':
                c = 13;
                break;
            case 'l':
                c = 10;
                break;
            case 't':
                c = 9;
                break;
            case 'b':
                c = 8;
                break;
            case 'f':
                c = 12;
                break;
            default:
                c = *(src - 1);
            }
            if (--size > 0) {
                *tok++ = c;
            }
        } else {
            /* check for the end of the token */
            if (quotef) {
                if (*src == '"')
                    break;
            } else {
                if (*src == ' ' || *src == '\t')
                    break;
            }

            /* set quote mode if quote found */
            if (*src == '"')
                quotef = TRUE;

            /* record the character */
            c = *src++;
            if (--size > 0)
                *tok++ = c;
        }
    }

    /* terminate the token and exit */
    if (*src)
        ++src;
    *tok = 0;
    return (src);
}



/*++

Used in .rhorc, for example, to store a procedure which is then bound
to a key.

++*/
int store_procedure(f, n)
int f, n;

{
    struct BUFFER *bp;          /* pointer to macro buffer */
    int status;                 /* return status */
    char bname[LENPATH];        /* name of buffer to use */

    f = n = 0;                  /* splint */

    /* get the name of the procedure */
    if ((status = mlreply(TEXT114, &bname[1], LENPATH - 2)) != TRUE)    /*"Procedure name: " */
        return (status);

    /* construct the macro buffer name */
    bname[0] = '[';
    if (!sourcing)
        strcat(bname, ")");
    else
        strcat(bname, "]");

    /* set up the new macro buffer */
    if ((bp = exist_buffer(bname, TRUE, BFINVS)) == NULL) {
        mlwrite(TEXT113);       /*"Can not create macro" */
        return (FALSE);
    }

    /* and make sure it is empty */
    clear_buffer(bp);

    /* and set the macro store pointers to it */
    mstore = TRUE;
    bstore = bp;
    return (TRUE);
}



/*++

Used by macros and procedures to execute another procedure.

++*/
int execute_procedure(f, n)
int f, n;

{
    BUFFER *bp;                 /* ptr to buffer to execute */
    int status;                 /* status return */
    char bufn[LENPATH + 2];     /* sourced buffer to execute? */
    char *bufn2;                /* rhorc buffer to execute?   */

    f = 0;                      /* splint */

    /* find out what buffer the user wants to execute */
    if ((status = mlreply(TEXT115, &bufn[1], LENPATH)) != TRUE) /*"Execute procedure: " */
        return (status);

    /* construct the buffer name */
    bufn[0] = '[';
    strcat(bufn, "]");
    bufn2 = strdup(bufn);
    bufn2[strlen(bufn2) - 1] = ')';

    /* find the pointer to that buffer */
    if ((bp = exist_buffer(bufn, FALSE, 0)) == NULL) {
        if ((bp = exist_buffer(bufn2, FALSE, 0)) == NULL) {
            mlwrite(TEXT116);   /* "No such procedure" */
            return (FALSE);
        }
    }

    /* and now execute it as asked */
    while (n-- > 0)
        if ((status = do_buffer(bp)) != TRUE)
            return (status);
    return (TRUE);
}




/*++

Send a buffer to the macro engine.

++*/
int execute_buffer(f, n)
int f, n;

{
    BUFFER *bp;                 /* ptr to buffer to execute */
    int status;                 /* status return */

    f = 0;                      /* splint */

    /* find out what buffer the user wants to execute */
    if ((bp = prompt4buffer(TEXT117, curbp->b_name, FALSE)) == NULL)    /* "Execute buffer: " */
        return (ABORT);

    /* and now execute it as asked */
    while (n-- > 0)
        if ((status = do_buffer(bp)) != TRUE)
            return (status);
    return (TRUE);
}



/*++

Expand all "%" to "%%"

++*/
void make_literal(s)
char *s;                        /* string to expand */

{
    char *sp;                   /* temp for expanding string */
    char *ep;                   /* ptr to end of string to expand */

    sp = s;
    while (*sp)
        if (*sp++ == '%') {
            /* advance to the end */
            ep = --sp;
            while (*ep++);
            /* null terminate the string one out */
            *(ep + 1) = 0;
            /* copy backwards */
            while (ep-- > sp)
                *(ep + 1) = *ep;

            /* and advance sp past the new % */
            sp += 2;
        }
}



/*++

Move file into a buffer and execute it. If there are no errors, delete
the buffer on exit.

++*/
int do_file(fname)
char *fname;                    /* file name to execute */

{
    BUFFER *bp;                 /* buffer to place file to execute */
    BUFFER *cb;                 /* temp to hold current buf while we read */
    int status;                 /* results of various calls */

    char *bname = "do_file";    /* avoid name clashes */
    if ((bp = exist_buffer(bname, TRUE, 0)) == NULL)
        return (FALSE);

    bp->b_mode = MDVIEW;        /* mark the buffer as read only */
    cb = curbp;                 /* save the old buffer */
    curbp = bp;                 /* make this one current */

    /* and try to read in the file to execute */
    if ((status = readin(fname, FALSE)) != TRUE) {
        curbp = cb;             /* restore the current buffer */
        return (status);
    }

    /* go execute it */
    curbp = cb;                 /* restore the current buffer */
    if ((status = do_buffer(bp)) != TRUE)
        return (status);

    /* if not displayed, remove buffer. */
    if (bp->b_nwnd == 0)
        release_buffer(bp);
    return (TRUE);
}



/*++

Load a macro file from ~/.rho

++*/
int source(f, n)
int f, n;

{
    int status;                 /* return status of name query */
    char fname[NSTRING];        /* name of file to execute */
    char srcfile[LENPATH];

    sourcing = TRUE;
    purge_sbuffers();

    f = 0;                      /* splint */

    if ((status = mlreply(TEXT129, fname, NSTRING - 1)) != TRUE)        /*"File to source: " */
        return (status);

    /* look up the path for the file */
    strcpy(srcfile, getenv("HOME"));
    strcat(srcfile, "/.rho/source.");
    strcat(srcfile, fname);

    if (!fexist(srcfile)) {
        /* complain if we are interactive */
        if (clexec == FALSE)
            mlwrite(TEXT214, srcfile);  /*"%%No such file as %s" */
        return (FALSE);
    }

    /* execute it */
    while (n-- > 0)
        if ((status = do_file(srcfile)) != TRUE)
            return (status);

    sourcing = FALSE;

    return (TRUE);
}


/*++

Take a passed string as a command line and translate it to be executed
as a command. This function will be used by execute-command-line and by
all source and startup files. Lastflag/thisflag is also updated.
Format of the command line is:

	{# arg} <command-name> {<argument string(s)>}

++*/
static int do_cmd(cline)
char *cline;                    /* command line to execute */

{

    BUFFER *bp;                 /* buffer to execute */
    char *oldestr;              /* original exec string */
    char bufn[LENPATH + 2];     /* sourced buffer to execute? */
    char *bufn2;                /* rhorc buffer to execute?   */
    char tkn[NSTRING];          /* next token off of command line */
    int (*fnc) ();              /* function to execute */
    int f;                      /* default argument flag */
    int n;                      /* numeric repeat value */
    int oldcle;                 /* old contents of clexec flag */
    int status;                 /* return status of function */

    /* if we are scanning and not executing, return */
    if (execlevel)
        return (TRUE);

    oldestr = execstr;          /* save last ptr to string to execute */
    execstr = cline;            /* and set this one as current        */

    /* first set up the default command values */
    f = FALSE;
    n = 1;
    lastflag = thisflag;
    thisflag = 0;

    /* and grab the first token */
    if ((status = get_macro_line(tkn)) != TRUE) {
        execstr = oldestr;
        return (status);
    }

    /* process leading argument */
    if (get_TKtype(tkn) != TKCMD) {
        f = TRUE;
        strcpy(tkn, fixnull(get_TKval(tkn)));
        n = str2int(tkn);

        /* get the command to execute */
        if ((status = get_macro_line(tkn)) != TRUE) {
            execstr = oldestr;
            return (status);
        }
    }

    /* match the token to see if it exists */
    if ((fnc = get_fn_name(tkn)) == NULL) {

        /* construct the buffer name */
        strcpy(bufn, "[");
        strcat(bufn, tkn);
        strcat(bufn, "]");

        bufn2 = strdup(bufn);
        bufn2[strlen(bufn2) - 1] = ')';

        /* find the pointer to that buffer */
        if ((bp = exist_buffer(bufn, FALSE, 0)) == NULL) {
            if ((bp = exist_buffer(bufn2, FALSE, 0)) == NULL) {
                mlwrite(TEXT16);        /*"[No such Function]" */
                execstr = oldestr;
                return (FALSE);
            }
        }

        /* execute the buffer */
        oldcle = clexec;        /* save old clexec flag */
        clexec = TRUE;          /* in cline execution */
        while (n-- > 0)
            if ((status = do_buffer(bp)) != TRUE)
                break;
        cmdstatus = status;     /* save the status */
        clexec = oldcle;        /* restore clexec flag */
        execstr = oldestr;
        return (status);
    }


    oldcle = clexec;            /* save old clexec flag */
    clexec = TRUE;              /* in cline execution */
    status = (*fnc) (f, n);     /* call the function */
    cmdstatus = status;         /* save the status */
    clexec = oldcle;            /* restore clexec flag */
    execstr = oldestr;
    return (status);
}
